/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* created: 03/11/2008 
   updated: 30/11/2008 */

#ifndef SCESCENEENTITY_H
#define SCESCENEENTITY_H

#include <SCE/utils/SCEList.h>
#include <SCE/interface/SCENode.h>
#include <SCE/interface/SCEGeometryInstance.h>
#include <SCE/interface/SCELevelOfDetail.h>
#include <SCE/interface/SCESceneResource.h>
#include <SCE/interface/SCELight.h>
#include <SCE/interface/SCEBoundingBox.h>
#include <SCE/interface/SCECamera.h>

#ifdef __cplusplus
extern "C"
{
#endif

/** Maximum number of lights that can affect an instance */
#define SCE_MAX_LIGHTS_PER_INSTANCE 8

/** \copydoc sce_ssceneentityproperties */
typedef struct sce_ssceneentityproperties SCE_SSceneEntityProperties;
struct sce_ssceneentityproperties
{
    int cullface;
    int cullmode;
    int depthtest;
    int depthmode;
    int alphatest;
};

/** \copydoc sce_ssceneentityinstance */
typedef struct sce_ssceneentityinstance SCE_SSceneEntityInstance;
/** \copydoc sce_ssceneentity */
typedef struct sce_ssceneentity SCE_SSceneEntity;
/** \copydoc sce_ssceneentitygroup */
typedef struct sce_ssceneentitygroup SCE_SSceneEntityGroup;

/**
 * \brief An instance of a scene entity
 * \sa SCE_SSceneEntity
 */
struct sce_ssceneentityinstance
{
    SCE_SNode *node;                 /**< Node */
    SCE_SGeometryInstance *instance; /**< Geometry instance */
    SCE_SSceneEntityProperties props; /**< Instance's properties (GL states) */
#if 0
    /** Lights affecting the instance */
    SCE_SLight *lights[SCE_MAX_LIGHTS_PER_INSTANCE];
    unsigned n_lights;               /**< Number of lights in \c lights */
#endif
    SCE_SLevelOfDetail *lod;         /**< LOD managment structure */
    SCE_SSceneEntity *entity;        /**< Entity used by the instance */
    int selected;                    /**< Is it selected for rendering? */
    SCE_SSceneEntityGroup *group;    /**< Group that contains the instance */
    int removed;                     /**< Is it removed from its group? */
#if SCE_LIST_ITERATOR_NO_MALLOC
    SCE_SListIterator iterator;
#endif
    SCE_SListIterator *it;           /**< Own iterator for fast add/remove */
};

/**
 * \brief A scene entity that defines a model that can be rendered via an
 * instance
 * \sa SCE_SSceneEntityInstance
 */
struct sce_ssceneentity
{
    SCE_SGeometryInstanceGroup *igroup; /**< Geometry group */

    SCE_SMesh *mesh;              /**< Mesh of the entity */
    SCE_SBoundingBox box;         /**< \c mesh's bounding box */
    SCE_SList *textures;          /**< Textures used by the entity */
    SCE_SSceneResource *shader;   /**< Shader used by the entity */
    SCE_SSceneResource *material; /**< Material used by the entity */

    SCE_SSceneEntityGroup *group; /**< Group of the entity */
#if SCE_LIST_ITERATOR_NO_MALLOC
    SCE_SListIterator iterator;
#endif
    SCE_SListIterator *it;        /**< Own iterator for group */
};

/**
 * \brief Defines a group of entities they represent the same geometry but in
 * a different LOD
 */
struct sce_ssceneentitygroup
{
    SCE_SList *instances;       /**< Instances to update & manage */
    SCE_SList *entities;
    unsigned int n_entities;
};

/* callback prototype for 'ForEach' function */
typedef int (*SCE_FForEachSceneEntityInstanceFunc)(SCE_SSceneEntityGroup*,
                                                   SCE_SSceneEntityInstance*,
                                                   void*);

void SCE_SceneEntity_InitProperties (SCE_SSceneEntityProperties*);

void SCE_SceneEntity_InitInstance (SCE_SSceneEntityInstance*);
SCE_SSceneEntityInstance* SCE_SceneEntity_CreateInstance (void);
void SCE_SceneEntity_DeleteInstance (SCE_SSceneEntityInstance*);

void SCE_SceneEntity_Init (SCE_SSceneEntity*);
SCE_SSceneEntity* SCE_SceneEntity_Create (void);
void SCE_SceneEntity_Delete (SCE_SSceneEntity*);

void SCE_SceneEntity_InitGroup (SCE_SSceneEntityGroup*);
SCE_SSceneEntityGroup* SCE_SceneEntity_CreateGroup (void);
void SCE_SceneEntity_DeleteGroup (SCE_SSceneEntityGroup*);


void SCE_SceneEntity_AddEntity (SCE_SSceneEntityGroup*, SCE_SSceneEntity*);
void SCE_SceneEntity_RemoveEntity (SCE_SSceneEntity*);
#if 0
void SCE_SceneEntity_SortGroup (SCE_SSceneEntityGroup*);
#endif
unsigned int SCE_SceneEntity_GetGroupNumEntities (SCE_SSceneEntityGroup*);
SCE_SList* SCE_SceneEntity_GetGroupEntitiesList (SCE_SSceneEntityGroup*);

void SCE_SceneEntity_AddInstance (SCE_SSceneEntityGroup*,
                                  SCE_SSceneEntityInstance*);
void SCE_SceneEntity_RemoveInstance (SCE_SSceneEntityInstance*);

void SCE_SceneEntity_AddInstanceToEntity (SCE_SSceneEntity*,
                                          SCE_SSceneEntityInstance*);
void SCE_SceneEntity_RemoveInstanceFromEntity (SCE_SSceneEntityInstance*);


void SCE_SceneEntity_ForEachInstance (
    SCE_SSceneEntityGroup*, SCE_FForEachSceneEntityInstanceFunc, void*);
void SCE_SceneEntity_ForEachInstanceInEntity (
    SCE_SSceneEntity*, SCE_FForEachSceneEntityInstanceFunc, void*);

#if 0
void SCE_SceneEntity_AddInstanceLight (SCE_SSceneEntityInstance*, SCE_SLight*);
void SCE_SceneEntity_ClearInstanceLights (SCE_SSceneEntityInstance*);

SCE_SLight** SCE_SceneEntity_GetInstanceLights (SCE_SSceneEntityInstance*);
unsigned int SCE_SceneEntity_GetInstanceNumLights (SCE_SSceneEntityInstance*);
#endif

SCE_SNode* SCE_SceneEntity_GetInstanceNode (SCE_SSceneEntityInstance*);
SCE_SGeometryInstance*
SCE_SceneEntity_GetInstanceInstance (SCE_SSceneEntityInstance*);
SCE_SSceneEntityProperties*
SCE_SceneEntity_GetInstanceProperties (SCE_SSceneEntityInstance*);
SCE_SLevelOfDetail*
SCE_SceneEntity_GetInstanceLOD (SCE_SSceneEntityInstance*);
int SCE_SceneEntity_IsInstanceSelected (SCE_SSceneEntityInstance*);


void SCE_SceneEntity_SetMesh (SCE_SSceneEntity*, SCE_SMesh*);
SCE_SMesh* SCE_SceneEntity_GetMesh (SCE_SSceneEntity*);

int SCE_SceneEntity_AddTexture (SCE_SSceneEntity*, SCE_SSceneResource*);
void SCE_SceneEntity_RemoveTexture (SCE_SSceneEntity*, SCE_SSceneResource*);
SCE_SList* SCE_SceneEntity_GetTexturesList (SCE_SSceneEntity*);

int SCE_SceneEntity_SetShader (SCE_SSceneEntity*, SCE_SSceneResource*);
SCE_SSceneResource* SCE_SceneEntity_GetShader (SCE_SSceneEntity*);

int SCE_SceneEntity_SetMaterial (SCE_SSceneEntity*, SCE_SSceneResource*);
SCE_SSceneResource* SCE_SceneEntity_GetMaterial (SCE_SSceneEntity*);

int SCE_SceneEntity_HaveResourceOfType (SCE_SSceneEntity*, int);
int SCE_SceneEntity_HaveInstance (SCE_SSceneEntity*);


void SCE_SceneEntity_DetermineInstanceLOD (SCE_SSceneEntityInstance*,
                                         SCE_SCamera*);
void SCE_SceneEntity_DetermineLODs (SCE_SSceneEntityGroup*, SCE_SCamera*);

int SCE_SceneEntity_IsInstanceInFrustum (SCE_SSceneEntityInstance*,
                                         SCE_SCamera*);
void SCE_SceneEntity_SelectInFrustumInstances (SCE_SSceneEntityGroup*,
                                               SCE_SCamera*, int);

void SCE_SceneEntity_ApplyProperties (SCE_SSceneEntityInstance*);

void SCE_SceneEntity_UseResources (SCE_SSceneEntity*);

void SCE_SceneEntity_Render (SCE_SSceneEntity*);
void SCE_SceneEntity_RenderGroup (SCE_SSceneEntityGroup*);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* guard */
